#include <linux/config.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/kernel.h>       /* printk() */
#include <linux/slab.h>         /* kmalloc() */
#include <linux/fs.h>           /* everything... */
#include <linux/errno.h>        /* error codes */
#include <linux/types.h>        /* size_t */
#include <linux/fcntl.h>        /* O_ACCMODE */
#include <linux/hdreg.h>        /* HDIO_GETGEO */
#include <asm/system.h>         /* cli(), *_flags */
#include <linux/ioctl.h>
#include <linux/vmalloc.h>
#include <linux/ts_sbc.h>
#include <linux/reboot.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include <asm/irq.h>

#define MAJOR_NR sdcard_major   /* force definitions on in blk.h */
int sdcard_major;               /* must be declared before including blk.h */

#include <linux/blk.h>
#include <linux/blkpg.h>        /* blk_ioctl() */
#include <linux/sched.h>

#include <asm/dma.h>

extern void __udelay(unsigned int);
#define udelay(x) __udelay((x))



#ifdef CONFIG_PROSOFT_SD
#include "tssdcore2.c"
#define TS7XXX_SDCARD1          0x600FF000
#else
#include "tssdcore2.c"
#define TS7XXX_SDCARD1	(TS7XXX_IO8_BASE_PHYS|SDCARD1_OFFSET)
#define SDCARD1_OFFSET	0x03000000
#endif

#define SD_DEFNAME		"sdcard0"
#define SD_DEFIOSIZE	4
#define SD_NUMDEV		1
#define SD_SHIFT		4	//max 16 partitions
#define SD_NUMPART		1 << 4
#define SD_RAHEAD		0
#define SD_BLKSIZE		1024  
#define SD_HARDSECT		512  

#define DMA_ENABLE		1
#define DMA_CHANNEL		10
#define DMA_BUFSIZE		1024
#define DMA_M2M_0_SIZE	0x40
#define DMA_M2M_1_SIZE	0x40

static unsigned char *name = SD_DEFNAME;
static unsigned int io = TS7XXX_SDCARD1;
static int rahead = SD_RAHEAD;
static int hardsect = SD_HARDSECT;
static int dmaenable = 0;
static int dmach = 1;
static int sd_blocksizes[1 << SD_SHIFT];

MODULE_AUTHOR("Technologic Systems");
MODULE_LICENSE("Proprietary");
MODULE_PARM(io, "l");
MODULE_PARM(name, "s");
MODULE_PARM(rahead, "i");
MODULE_PARM(dmach, "l");
MODULE_PARM(dmaenable, "l");

struct sdcarddev {
	
	struct sdcore sc_sdcore;
	
	int sdcard_size;
	int sdcard_sizes[SD_NUMPART];
	int sdcard_hardsect_sizes[SD_NUMPART];
	struct hd_struct sdcard_partitions[SD_NUMPART];	
	struct gendisk sdcard_gendisk;
	devfs_handle_t devfs_handle;

	struct timer_list timer;
	spinlock_t lock;	

	int irq;
	unsigned long dmaregs;
	int dmach;

	int busy, thread, parksect;		
};

static struct sdcarddev *sc;
static struct block_device_operations sdcard_bdops;

DECLARE_WAIT_QUEUE_HEAD (qthr);
DECLARE_WAIT_QUEUE_HEAD (qdma);


#define DMA_DECLBASE(x)	unsigned int dmaregs = (x);
#define DMA_WRITE(x, y)	*(volatile uint32_t *)(dmaregs + M2M_OFFSET_ ## x) = (y) 
#define DMA_READ(x) *(volatile uint32_t *)(dmaregs + M2M_OFFSET_ ## x)


static void int_handler(int irq, void *dev_id, struct pt_regs * regs)
{
	struct sdcarddev *scarg = (struct sdcarddev *)dev_id;
	DMA_DECLBASE(scarg->dmaregs);

	DMA_WRITE(INTERRUPT, 0x0);
	wake_up_interruptible(&qdma);	
	
	return;
}

pte_t * vmalloc_to_pte(void * vmalloc_addr)
{
	unsigned long addr = (unsigned long) vmalloc_addr;
	volatile int *v = *(int *)vmalloc_addr;
	pmd_t *pmd;
	pte_t *pte = NULL;
	pgd_t *pgd;
	
	pgd = pgd_offset_k(addr);
	if (!pgd_none(*pgd)) {
		pmd = pmd_offset(pgd, addr);
		if (!pmd_none(*pmd)) {
			pte = pte_offset(pmd, addr);
		}
	}
	return pte;
}


static int sdcard_dmastream(void *os_arg, unsigned char *dat, unsigned int buflen)
{
	struct sdcarddev *scarg = (struct sdcarddev *)os_arg;
	int i;
	unsigned int s;
	DMA_DECLBASE(scarg->dmaregs);
	static unsigned char dummymem[8];
	static unsigned long dummymem_paddr = 0;
	static unsigned char bufmem[520];
	
	//udelay(2000);
	//udelay(10000);

	// s = inl(SMCBCR1);
	//outl(0x8602, SMCBCR1);
	DMA_WRITE(CONTROL, 0x11085000);
	DMA_WRITE(SAR_BASE0, io|SDDAT );	
	
	if (dat) {
		// pte_t *pte = vmalloc_to_pte(bufmem);
		//unsigned long paddr = pte_pfn(*pte) << PAGE_SHIFT;
		/* paddr += (unsigned long)bufmem & 0xfff;
		*/

		// DMA_WRITE(DAR_BASE0, virt_to_bus(dat));
		/*
		printk("dat:%x, paddr:%x, *pte:%x, pte:%x\n", dat, paddr, *pte, pte);
		printk("virt_to_bus:%x\n", virt_to_bus(dat));
		*/
		//printk("%x=>%x\n", dat, virt_to_bus(dat));
		//paddr = virt_to_bus(dat);
		DMA_WRITE(DAR_BASE0, virt_to_bus(dat));
		}
	else {
		/*
		struct page *pg = vmalloc_to_page(dummymem);
		unsigned long pfn = page_to_pfn(pg);
		unsigned int node = PFN_TO_NID(pfn);
		struct pglist_data *nd = NODE_DATA(node);
		unsigned long paddr = nd->node_start_paddr;
		unsigned long off = pg - NODE_MEM_MAP(node);
		paddr += (off << PAGE_SHIFT) + ((unsigned long)dummymem & 0xfff);
		*/
		if (dummymem_paddr == 0) {
			pte_t *pte = vmalloc_to_pte(dummymem);
			dummymem_paddr = pte_pfn(*pte) << PAGE_SHIFT;
			dummymem_paddr += (unsigned long)dummymem & 0xfff;
		}
			
		DMA_WRITE(DAR_BASE0, dummymem_paddr);
	}		
	
	DMA_WRITE(BCR0, buflen);
	
//	if (buflen > 8) {
//		DMA_WRITE(CONTROL, 0x1108500c);
//		interruptible_sleep_on(&qdma);
//		if (signal_pending (current)) 
//			return 1;				
//	} else {
		DMA_WRITE(CONTROL, 0x11085008);
		while (!(DMA_READ(STATUS) & 0x40));
		DMA_WRITE(INTERRUPT, 0x0);
//	}
	// outl(s, SMCBCR1);
	/* printk("--\n");
	for (i = 0; i < (buflen / 16); i++) {
		int j;
		for (j = 0; j < 16; j++) {
			printk("%02x ", bufmem[(i * 16) + j]);
		}
		printk("\n");
	}
	*/
	
	return 0;
}


static int sdcard_dmaprep(void *os_arg, unsigned char *buf, unsigned int buflen)
{
	invalidate_dcache_range(buf, buf + buflen);
	return 0;
}


static void sdcard_delay(void *arg, unsigned int us)
{
	udelay(us);
}


static int sdcard_open(struct inode *inode, struct file *filp)
{
    MOD_INC_USE_COUNT;
    return 0;
}


static int sdcard_release(struct inode *inode, struct file *filp)
{
    MOD_DEC_USE_COUNT;
    return 0;
}


static int sdcard_ioctl(struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg)
{
    int err;
    struct hd_geometry geo;

    switch (cmd) {
    case HDIO_GETGEO:
        /*
         * get geometry: we have to fake one...  trim the size to a
         * multiple of 512 (256k): tell we have 32 sectors, 16 heads,
         * whatever cylinders.
         */

        // TODO: This should match what the SD cards look like in a 
        // USB adapter dongle.  -- JO

        err = !access_ok(VERIFY_WRITE, arg, sizeof(geo));
        if (err)
            return -EFAULT;

        geo.cylinders = sc->sdcard_size / (16 * 32);
        geo.heads = 16;
        geo.sectors = 32;
        geo.start = 0;

        if (copy_to_user((void *) arg, &geo, sizeof(geo)))
            return -EFAULT;

        return 0;

    default:
        /*
         * For ioctls we don't understand, let the block layer handle them.
         */
        return blk_ioctl(inode->i_rdev, cmd, arg);
    }

    return -ENOTTY;
}


static int sdcard_check_change(kdev_t i_rdev)
{
	return 0;
}



static int sdcard_revalidate(kdev_t i_rdev)
{
    /* first clear old partition information */
    memset(sc->sdcard_gendisk.sizes+1, 0, (sc->sdcard_gendisk.max_p-1)*sizeof(int));
    memset(sc->sdcard_gendisk.part+1, 0, (sc->sdcard_gendisk.max_p-1)*sizeof(struct hd_struct));

    sc->sdcard_size = sc->sdcard_gendisk.part[0].nr_sects = sdreset(&sc->sc_sdcore);

    /* then fill new info */
    register_disk(&sc->sdcard_gendisk, i_rdev, (sc->sdcard_gendisk.max_p), &sdcard_bdops, sc->sdcard_size);

    return 0;
}


static int sdcard_make_request(request_queue_t * q, int rw, struct buffer_head *bh)
{
	unsigned int ret = 0, sect;
	static int busy = 0;

	if (busy++ != 0) panic("recursive make_request!\n");

    sect = bh->b_rsector + sc->sdcard_partitions[MINOR(bh->b_rdev)].start_sect;
    del_timer(&sc->timer);

    switch (rw) {
    case READ:
    case READA:
        ret = sdread(&sc->sc_sdcore, sect, bh->b_data, bh->b_size / hardsect);
		if (ret) {
			sdreset(&sc->sc_sdcore);
            ret = sdread(&sc->sc_sdcore, sect, bh->b_data, bh->b_size / hardsect);
		}
		//printk(KERN_INFO "r %db %x %x\n", bh->b_size, sc->sdcard_partitions[MINOR(bh->b_rdev)].start_sect, bh->b_rsector);
        bh->b_end_io(bh, !ret);
	sc->parksect = sect + (bh->b_size / hardsect) - 1;
        break;
        
    case WRITE:
	//refile_buffer(bh);
        ret = sdwrite(&sc->sc_sdcore, sect, bh->b_data, bh->b_size / hardsect);
		if (ret && !sc->sc_sdcore.sd_wprot) {
			sdreset(&sc->sc_sdcore);
       	    ret = sdwrite(&sc->sc_sdcore, sect, bh->b_data, bh->b_size / hardsect);
		}
		//printk(KERN_INFO "w %db %x %x\n", bh->b_size, sc->sdcard_partitions[MINOR(bh->b_rdev)].start_sect, bh->b_rsector);
		if (ret == 0) mark_buffer_uptodate(bh, 1);
    	sc->timer.expires = jiffies + (HZ / 2);
    	add_timer(&sc->timer);
        bh->b_end_io(bh, !ret);
        break;
        
    default:
        bh->b_end_io(bh, 0);
        break;
    }
    busy--;
    return 0;
}


static int sdcard_transfer(const struct request *req)
{
    unsigned int ret = 0;
    int size;
    u8 *ptr;
    
    del_timer(&sc->timer);
    ptr = sc->sdcard_partitions[MINOR(req->rq_dev)].start_sect + req->sector;
    size = req->current_nr_sectors * hardsect;

    switch(req->cmd) {
        case READ:
        	ret = sdread(&sc->sc_sdcore, ptr, req->buffer, req->current_nr_sectors);
			if (ret) {
				sdreset(&sc->sc_sdcore);
    	        ret = sdread(&sc->sc_sdcore, ptr, req->buffer, req->current_nr_sectors);
		    }            
			//printk(KERN_INFO "r%db,%x,%d\n", size, ptr, req->current_nr_sectors);
		sc->parksect = ptr + req->current_nr_sectors - 1;
            return !ret;
        case WRITE:
			//sc->sc_sdcore.sd_erasehint = req->current_nr_sectors;
	        ret = sdwrite(&sc->sc_sdcore, ptr, req->buffer, req->current_nr_sectors);
			if (ret && !sc->sc_sdcore.sd_wprot) {
				sdreset(&sc->sc_sdcore);
				// sc->sc_sdcore.sd_erasehint = req->current_nr_sectors;
	       	    ret = sdwrite(&sc->sc_sdcore, ptr, req->buffer, req->current_nr_sectors);
			}
			//printk(KERN_INFO "w%db,%x,%d\n", size, ptr, req->current_nr_sectors);
    	sc->timer.expires = jiffies + HZ;
    	add_timer(&sc->timer);
            return !ret;
        default:
            return 0;
    }   
}

static void sdcard_processrequest(request_queue_t *q) {
	struct request *req;
	int status;
	
	spin_lock_irq (&io_request_lock);		
		
	while(! list_empty(&q->queue_head)) {
			
		req = blkdev_entry_next_request(&q->queue_head);
		blkdev_dequeue_request(req);
		spin_unlock_irq (&io_request_lock); 
			
        //printk(KERN_INFO "request thread %p: cmd %i sec %li (nr. %li)\n", req, req->cmd, req->sector, req->current_nr_sectors);		
		
		spin_lock(&sc->lock);
		do {
			status = sdcard_transfer(req);
		} while (end_that_request_first(req, status, name));		
		spin_unlock(&sc->lock);

		spin_lock_irq (&io_request_lock);
		end_that_request_last(req);  
	}
		
	spin_unlock_irq (&io_request_lock); 			
}

static int sdcard_thread(void *data)
{
	daemonize();
	exit_files(current);
	reparent_to_init();
	sprintf(current->comm, name);
	
	sc->busy = 0;
	sc->thread = 1;
	
	for (;;) {
		interruptible_sleep_on(&qthr);
		if (signal_pending (current)) 
			return 0;
			
		sdcard_processrequest(&blk_dev[sdcard_major].request_queue);
		
		spin_lock(&sc->lock);
		sc->busy = 0;
		spin_unlock(&sc->lock);
	}
	
	return 0;	
}


static void sdcard_request(request_queue_t *q)
{
	if (sc->busy)
		return;
		
	spin_lock(&sc->lock);
	sc->busy = 1;
	spin_unlock(&sc->lock);
	
	if (!sc->thread) {
		sdcard_processrequest(q);
	    
		spin_lock(&sc->lock);
		sc->busy = 0;
		spin_unlock(&sc->lock);
	}
	
	else {
		wake_up_interruptible(&qthr);	
	}
	
}

static struct block_device_operations sdcard_bdops = {
  open:					sdcard_open,
  release:				sdcard_release,
  ioctl:				sdcard_ioctl,
  revalidate:			sdcard_revalidate,
  check_media_change:	sdcard_check_change,
};

static void sdcommit(unsigned long *ptr) {
	char buf[512];

	sdread(&sc->sc_sdcore, sc->parksect, buf, 1); 
}

static int shutdownhook(struct notifier_block *NotifierBlock,
                           unsigned long Event,
                           void *Buffer)
{
  del_timer(&sc->timer);
  if (!(Event == SYS_RESTART || Event == SYS_HALT || Event == SYS_POWER_OFF))
    return NOTIFY_DONE;
  sdcommit(NULL);
  return NOTIFY_OK;
}

static struct notifier_block shutdownhook_notifier = { shutdownhook, NULL, 0 };

static int sdcard_init(void)
{
    unsigned int scratchreg0, i; 
   	
    /*
    int model = inw(TS7XXX_MODEL) & 0x7;
    int pldrev = inw(TS7XXX_PLDREV) & 0x7;
    int opts = inw(0x02400000 + TS7XXX_IO16_BASE) & 0x7;
    if (!((model == TS7260 && pldrev >= 5 && (opts & 0x4)) || model == TS7300)) {
    	printk("%s: no SD card option present on this board\n", name);
	return 1;
    }
    */
    
    printk("%s: Technologic Systems SD card controller, address 0x%x\n", name, io);
    
    sc = kmalloc(sizeof(struct sdcarddev), GFP_KERNEL);
	memset(sc, 0, sizeof(struct sdcarddev));
    
    
    spin_lock_init(&sc->lock);
    //kernel_thread(sdcard_thread, NULL, CLONE_FS | CLONE_FILES | CLONE_SIGHAND);
    //kernel_thread(sdcard_thread, NULL, 0);    
    
    
	/*** DMA INIT ***/ 
    sc->irq = IRQ_DMAM2M1;
    sc->dmach = DMA_CHANNEL;
    
	if (dmach) 
	  sc->dmaregs = (unsigned long) __ioremap(DMA_M2M_1_BASE, DMA_M2M_1_SIZE, 0);
	else
	  sc->dmaregs = (unsigned long) __ioremap(DMA_M2M_0_BASE, DMA_M2M_0_SIZE, 0);
	
	if (dmaenable) {
		// request_irq(sc->irq, int_handler, SA_SHIRQ, name, sc);
		sc->sc_sdcore.os_dmastream = sdcard_dmastream;
		sc->sc_sdcore.os_dmaprep = sdcard_dmaprep;		

	// Enable M2M DMA channel 1 power
	outl(0xaa, SYSCON_SWLOCK);
	if (dmach)
	  outl(inl(SYSCON_PWRCNT)|SYSCON_PWRCNT_DMA_M2MCH1, SYSCON_PWRCNT);
	else
	  outl(inl(SYSCON_PWRCNT)|SYSCON_PWRCNT_DMA_M2MCH0, SYSCON_PWRCNT);

	// Enable M2M DMA channel 1 DRQ pins
	outl(0xaa, SYSCON_SWLOCK);
	if (dmach)
	  outl(inl(SYSCON_DEVCFG)|SYSCON_DEVCFG_D1onG, SYSCON_DEVCFG); 
	else
	  outl(inl(SYSCON_DEVCFG)|SYSCON_DEVCFG_D0onG, SYSCON_DEVCFG); 

	// Speed up bus timing -- potentially breaks ISA bus cycles! ///?
	//outl(0x8601, SMCBCR1);	
	
	}
	else {
		sc->sc_sdcore.os_dmastream = NULL;
		sc->sc_sdcore.os_dmaprep = NULL;		
	}

	sc->busy = 0;
	sc->thread = 0;    
	sc->sc_sdcore.os_arg = sc;
	sc->sc_sdcore.os_delay = sdcard_delay;
	sc->sc_sdcore.sd_writeparking = 1;
	
	//sc->sc_sdcore.sd_regstart = (unsigned long) IO_ADDRESS(TS7XXX_IO8_BASE|SDCARD1_OFFSET);
	sc->sc_sdcore.sd_regstart = (unsigned long) __ioremap(io, SD_DEFIOSIZE, 0);
	
	scratchreg0 = inl(SYSCON_SCRREG0); 
	if (scratchreg0) {
		sc->sc_sdcore.sdboot_token = scratchreg0;
		outl(0, SYSCON_SCRREG0);
	}

	sc->sdcard_size = sdreset(&sc->sc_sdcore);
    if (sc->sdcard_size == 0) {
    	printk("%s: no card found\n", name);
    	kfree(sc);
		return 1;
    } else {	    
    	printk("%s: card size %d sectors\n", name, sc->sdcard_size);
    }

    init_timer(&sc->timer);
    sc->timer.function = sdcommit;
    sc->parksect = 0;
    register_reboot_notifier(&shutdownhook_notifier);
	
    sdcard_major = register_blkdev(0, name, &sdcard_bdops);
    sc->devfs_handle = devfs_mk_dir(NULL, name, NULL);

    sc->sdcard_gendisk.minor_shift = SD_SHIFT;
    sc->sdcard_gendisk.max_p = SD_NUMPART;
    sc->sdcard_gendisk.fops = &sdcard_bdops;
    sc->sdcard_gendisk.major = sdcard_major;
    sc->sdcard_gendisk.major_name = name;
    sc->sdcard_gendisk.nr_real = SD_NUMDEV;

    for(i = 0; i < SD_NUMPART; i++)
    	sc->sdcard_hardsect_sizes[i] = 512; // sc->sc_sdcore.sd_blocksize;
    memset(sc->sdcard_sizes, 0, sc->sdcard_gendisk.max_p*sizeof(int) );
    memset(sc->sdcard_partitions, 0, sc->sdcard_gendisk.max_p*sizeof(struct hd_struct) );
    sc->sdcard_sizes[0] = sc->sdcard_partitions[0].nr_sects = sc->sdcard_size;
    sc->sdcard_gendisk.sizes = sc->sdcard_sizes;
    sc->sdcard_gendisk.part = sc->sdcard_partitions;

    add_gendisk(&sc->sdcard_gendisk);

    blk_size[sdcard_major] = sc->sdcard_sizes;
    hardsect_size[sdcard_major] = sc->sdcard_hardsect_sizes;

    for(i=0;i<(1 << SD_SHIFT);i++)
	sd_blocksizes[i] = SD_BLKSIZE;

    blksize_size[sdcard_major] = sd_blocksizes;
    read_ahead[sdcard_major] = rahead;

    blk_queue_make_request(BLK_DEFAULT_QUEUE(sdcard_major), sdcard_make_request);
//	blk_init_queue(BLK_DEFAULT_QUEUE(sdcard_major), sdcard_request);

    register_disk(&sc->sdcard_gendisk, MKDEV(sdcard_major, 0), sc->sdcard_gendisk.max_p, &sdcard_bdops, sc->sdcard_size);        
    
    EXPORT_NO_SYMBOLS;

    return 0;
}


static void sdcard_cleanup(void)
{
    fsync_dev(MKDEV(sdcard_major, 0));
    devfs_register_partitions(&sc->sdcard_gendisk, 0, 1); //unregister
    unregister_blkdev(sdcard_major, name);

    del_gendisk(&sc->sdcard_gendisk);
    blk_size[sdcard_major] = NULL;
    
    free_irq(sc->irq, name);
    __iounmap((unsigned long *) sc->sc_sdcore.sd_regstart);
    __iounmap((unsigned long *) sc->dmaregs);
    __iounmap((unsigned char *) io);

    devfs_unregister(sc->devfs_handle);
    
    kfree(sc);
}


int init_module(void)
{
    return sdcard_init();
}


void cleanup_module(void)
{
    return sdcard_cleanup();
}


