/*
   Copyright (C) 2007 Jesse Off
   
   Based on i2c-parport-light.c
   Copyright (C) 2003-2004 Jean Delvare <khali@linux-fr.org>
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * ------------------------------------------------------------------------ */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/init.h>
#include <linux/i2c.h>
#include <linux/smp_lock.h>
#include <linux/pci.h>
#include <asm/io.h>
#include <linux/ioport.h>
#include <linux/delay.h>
#include <linux/i2c.h>
#include <linux/i2c-algo-bit.h>

#define EE_DELAY_USEC       50
#define GPIOG_EECLK 1
#define GPIOG_EEDAT 2

static void ep93xx_setscl(void *data, int state)
{
	unsigned int val, dir;
	val = inl(GPIO_PGDR);
	dir = inl(GPIO_PGDDR);

	if (state) {
		dir &= ~GPIOG_EECLK;
		outl(dir, GPIO_PGDDR);
	} else {
		dir |= GPIOG_EECLK;
		val &= ~GPIOG_EECLK;
		outl(val, GPIO_PGDR);
		outl(dir, GPIO_PGDDR);
	}
}

static void ep93xx_setsda(void *data, int state)
{
	unsigned int val, dir;
	val = inl(GPIO_PGDR);
	dir = inl(GPIO_PGDDR);

	if (state) {
		dir &= ~GPIOG_EEDAT;
		outl(dir, GPIO_PGDDR);
	} else {
		dir |= GPIOG_EEDAT;
		val &= ~GPIOG_EEDAT;
		outl(val, GPIO_PGDR);
		outl(dir, GPIO_PGDDR);
	}
}

static int ep93xx_getscl(void *data)
{
	return (inl(GPIO_PGDR) & GPIOG_EECLK) ? 1 : 0;
}

static int ep93xx_getsda(void *data)
{
	return (inl(GPIO_PGDR) & GPIOG_EEDAT) ? 1 : 0;
}

static struct i2c_algo_bit_data ep93xx_algo_data = {
	.setsda = ep93xx_setsda,
	.setscl = ep93xx_setscl,
	.getsda = ep93xx_getsda,
	.getscl = ep93xx_getscl,
	.udelay = EE_DELAY_USEC,
	.mdelay = EE_DELAY_USEC * 1000,
	.timeout = HZ,
};

struct i2c_adapter ep93xx_i2c_adapter = {
	.name = "EP93XX I2C interface",
	.id = I2C_HW_B_LP, /* This is wrong */
	.algo_data = &ep93xx_algo_data,
};

static int __init i2c_ep93xx_init(void)
{
	ep93xx_setsda(NULL, 1);
	ep93xx_setscl(NULL, 1);

	if (i2c_bit_add_bus(&ep93xx_i2c_adapter) < 0) {
		printk(KERN_ERR "i2c-ep93xx: Unable to register with I2C\n");
		return -ENODEV;
	}

	return 0;
}

static void __exit i2c_ep93xx_exit(void)
{
	i2c_bit_del_bus(&ep93xx_i2c_adapter);
}

MODULE_AUTHOR("Jesse Off");
MODULE_DESCRIPTION("I2C bit banger for Cirrus EP93xx processors");
MODULE_LICENSE("GPL");

EXPORT_SYMBOL(ep93xx_i2c_adapter);

module_init(i2c_ep93xx_init);
module_exit(i2c_ep93xx_exit);

