/*
 *  linux/drivers/serial/tsuart-73.c
 *
 *  TS-UART loader for Technologic Systems TS-7300
 *
 *  (c) 2006 Technologic Systems
 *
 * This source code is hereby licensed to the Linux 2.4 kernel under the
 * GPL, and to Technologic Systems customers under the BSD license.
 *
 * 2006-05-12:mos:initial version completed
 *
 */
#include <linux/version.h>
#include <linux/slab.h>
#include <linux/config.h>
#include <linux/module.h>
#include <linux/tty.h>
#include <linux/tty_flip.h>
#include <linux/ioport.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/serial.h>
#include <linux/console.h>
#include <linux/sysrq.h>
#include <linux/proc_fs.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <linux/serial_core.h>
//#include <serial_core.h>
#include <asm/hardware/serial_amba.h>
#include <linux/ts_sbc.h>
#include <asm/uaccess.h>
#include "tsuart.h"

struct uart_driver *dr8=0,*dr9=0;

#define BOARD_ID_7300 0x7300

static int tryInitPort(int base) {
  struct tsuart_port *ts1,*ts2;

  ts1 = kmalloc(sizeof(struct tsuart_port),GFP_KERNEL);
  if (!ts1) {
    printk("Unable to allocate memory for tsuart port\n");
    return 0;
  }
  init_tsuart_port(ts1);
  ts1->rxsize = 1;
  ts1->u.iotype = SERIAL_IO_MEM;
  ts1->get = tsuartGetMEM16;
  ts1->put = tsuartPutMEM16;
  ts1->debugFlags = portDebug;
  ts1->u.mapbase = 0x72000000 + base;
  ts1->portStatus = &ts1->pstatus;
  ts1->pstatus = 0;
  ts1->u.irq = 40;
  ts1->mdmctrl = 0;
  ts1->localStatus |= PORT_STATUS_SHIRQ;
  ts1->boardId = BOARD_ID_7300;

  ts2 = kmalloc(sizeof(struct tsuart_port),GFP_KERNEL);
  if (!ts2) {
    printk("Unable to allocate memory for tsuart port\n");
    return 0;
  }
  init_tsuart_port(ts2);
  ts2->rxsize = 2;
  ts2->dataMask = 0x1FF;
  ts2->u.iotype = SERIAL_IO_MEM;
  ts2->get = tsuartGetMEM16_9;
  ts2->put = tsuartPutMEM16_9;
  ts2->debugFlags = portDebug;
  ts2->u.mapbase = 0x72000000 + base;
  ts2->portStatus = &ts1->pstatus;
  ts2->pstatus |= PORT_STATUS_TXMSB;
  ts2->u.irq = 40;
  ts2->mdmctrl = 0;
  ts2->localStatus |= PORT_STATUS_SHIRQ;
  ts2->boardId = BOARD_ID_7300;
  printk("registering %p,%p\n",ts1,ts2);
  if (tsuart_register_port(dr8,ts1) == 0) {
    //    return 1;
  }
  if (tsuart_register_port(dr9,ts2) == 0) {
    //    return 1;
  }
  return 1;
}

#define TTY_NAME8 "ttyT8S"
#define CUA_NAME8 "cuaT8S"
#define TTY_NAME9 "ttyT9S"
#define CUA_NAME9 "cuaT9S"

static char *tty = TTY_NAME8,*cua = CUA_NAME8;
static char *tty9 = TTY_NAME9,*cua9 = CUA_NAME9;
MODULE_PARM(tty,"s");
MODULE_PARM(cua,"s");
MODULE_PARM(tty9,"s");
MODULE_PARM(cua9,"s");

static int __init tsuart_init(void)
{
  int ports=0;
  int i;

  if (!tty) {
    tty = TTY_NAME8;
  }
  if (!cua) {
    cua = CUA_NAME8;
  }

  if (get_ts_model() != TS_MODEL_7300) {
    printk("TS-UART/7300 did not detect a TS-7300 board\n");
    return -1;
  } else {
    printk("TS-UART/7300 detected a TS-7300 board\n");
  }

  dr8 = kmalloc(sizeof(struct uart_driver),GFP_KERNEL);
  if (!dr8) {
    printk("Unable to allocate memory for tsuart port\n");
    return 0;
  }
  dr9 = kmalloc(sizeof(struct uart_driver),GFP_KERNEL);
  if (!dr9) {
    kfree(dr8);
    printk("Unable to allocate memory for tsuart port\n");
    return 0;
  }
 if (!tsuart_configure_driver(dr8,234,64,tty,cua,8)) {
    kfree(dr8);
    kfree(dr9);
    return 0;    
  }
  if (!tsuart_configure_driver(dr9,234,72,tty9,cua9,8)) {
    kfree(dr8);
    kfree(dr9);
    return 0;    
  }
  for (i=0;i<=28;i+=4) {
    if (tryInitPort(i)) {
      ports++;
    }
  }
  return (ports > 0) ? 0 : -1;
}

static void __exit tsuart_exit(void)
{
  // This module can be safely unloaded after initialization is complete.
  // We allow module to load only so error conditions can be detected.
}

module_init(tsuart_init);
module_exit(tsuart_exit);

EXPORT_NO_SYMBOLS;
