/*
 *  linux/drivers/serial/tsuart-bat3.c
 *
 *  TS-UART loader for Technologic Systems MARCTEL TS-7300 bitstream
 *
 *  (c) 2007 Technologic Systems
 *
 * This source code is hereby licensed to the Linux 2.4 kernel under the
 * GPL, and to Technologic Systems customers under the BSD license.
 *
 * 2007-05-11:mos:initial version completed
 *
 */
#include <linux/version.h>
#include <linux/slab.h>
#include <linux/config.h>
#include <linux/module.h>
#include <linux/tty.h>
#include <linux/tty_flip.h>
#include <linux/ioport.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/serial.h>
#include <linux/console.h>
#include <linux/sysrq.h>
#include <linux/proc_fs.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <linux/serial_core.h>
//#include <serial_core.h>
#include <asm/hardware/serial_amba.h>
#include <linux/ts_sbc.h>
#include <asm/uaccess.h>
#include "tsuart.h"

struct uart_driver *dr8=0,*dr9=0;

#define BOARD_ID_MARCTEL 0x735

unsigned tsuartGetMEM16_BRK(struct tsuart_port *port,int offset) {
  unsigned value;
  static int inbreak = 0;

  value = *((volatile unsigned short *)port->u.membase + offset);
  if (offset == 0) {
    if (bit_RXBRK(value) && !inbreak) {
      inbreak = 1;
      port->u.icount.brk++;
      uart_handle_break(&port->u);
      if (port->u.info->tty) {
	*(port->u.info->tty->flip.flag_buf_ptr) = TTY_BREAK;
	port->u.info->tty->flip.flag_buf_ptr++;
	port->u.info->tty->flip.char_buf_ptr++;
	port->u.info->tty->flip.count++;
      }
      //      if (uart_handle_sysrq_char(port, ch, regs)) {
      //      } else {
      //}
      tty_flip_buffer_push(port->u.info->tty);
    }
  } else {
    inbreak = 0;
  }
  return value;
}

unsigned tsuartGetMEM16_9_BRK(struct tsuart_port *port,int offset) {
  unsigned val = *((volatile unsigned short *)port->u.membase + offset);

  if (offset == 1) {
    return val & 0x1FF;
  } else if (offset == 0) {
    if (bit_RXBRK(val)) {
      uart_handle_break(&port->u);
    }    
  }
  return val;
}

static int tryInitPort(int i) {
  struct tsuart_port *ts1,*ts2;

  ts1 = kmalloc(sizeof(struct tsuart_port),GFP_KERNEL);
  if (!ts1) {
    printk("Unable to allocate memory for tsuart port\n");
    return 0;
  }
  init_tsuart_port(ts1);
  ts1->u.iotype = SERIAL_IO_MEM;
  ts1->get = tsuartGetMEM16_BRK;
  ts1->put = tsuartPutMEM16;
  ts1->debugFlags = portDebug;
  ts1->u.mapbase = 0x72000400+i*4;
  ts1->portStatus = &ts1->pstatus;
  ts1->pstatus = 0;
  ts1->boardId = BOARD_ID_MARCTEL;
  ts1->u.irq = 40;
  ts1->mdmctrl = 0;
  ts1->localStatus |= PORT_STATUS_SHIRQ;

  ts2 = kmalloc(sizeof(struct tsuart_port),GFP_KERNEL);
  if (!ts2) {
    printk("Unable to allocate memory for tsuart port\n");
    return 0;
  }
  init_tsuart_port(ts2);
  ts2->rxsize = 2;
  ts2->dataMask = 0x1FF;
  ts2->u.iotype = SERIAL_IO_MEM;
  ts2->get = tsuartGetMEM16_9_BRK;
  ts2->put = tsuartPutMEM16_9;
  ts2->debugFlags = portDebug;
  ts2->u.mapbase = 0x72000400+i*4;
  ts2->portStatus = &ts1->pstatus;
  ts2->pstatus |= PORT_STATUS_TXMSB;
  ts2->boardId = BOARD_ID_MARCTEL;
  ts2->u.irq = 40;
  ts2->mdmctrl = 0;
  ts2->localStatus |= PORT_STATUS_SHIRQ;

  if (tsuart_register_port(dr8,ts1) == 0) {
    //return 1;
  }
  if (tsuart_register_port(dr9,ts2) == 0) {
    //return 1;
  }
  return 0;
}

#define TTY_NAME8 "ttyT8Sv"
#define CUA_NAME8 "cuaT8Sv"
#define TTY_NAME9 "ttyT9Sv"
#define CUA_NAME9 "cuaT9Sv"

static char *tty = TTY_NAME8,*cua = CUA_NAME8;
static char *tty9 = TTY_NAME9,*cua9 = CUA_NAME9;
MODULE_PARM(tty,"s");
MODULE_PARM(cua,"s");
MODULE_PARM(tty9,"s");
MODULE_PARM(cua9,"s");

static int __init tsuart_init(void)
{
  int i,ports=0;
  volatile unsigned short int *cr;

  if (!tty) {
    tty = TTY_NAME8;
  }
  if (!cua) {
    cua = CUA_NAME8;
  }

  cr = (volatile unsigned short *)((int) __ioremap(0x72000000, 4096,0) + 0x422);
  if (*cr != 0x4D2A) {
    printk("TS-UART/MARCTEL did not find MARCTEL board\n");
    __iounmap(cr);
    return -1;
  } else {
    printk("TS-UART/MARCTEL found MARCTEL board\n");
    __iounmap(cr);
  }

  dr8 = kmalloc(sizeof(struct uart_driver),GFP_KERNEL);
  if (!dr8) {
    printk("Unable to allocate memory for tsuart port\n");
    return 0;
  }
  dr9 = kmalloc(sizeof(struct uart_driver),GFP_KERNEL);
  if (!dr9) {
    kfree(dr8);
    printk("Unable to allocate memory for tsuart port\n");
    return 0;
  }
 if (!tsuart_configure_driver(dr8,234,128,tty,cua,4)) {
    kfree(dr8);
    kfree(dr9);
    return 0;
  }
  if (!tsuart_configure_driver(dr9,234,132,tty9,cua9,4)) {
    kfree(dr8);
    kfree(dr9);
    return 0;
  }
  for (i=0;i<4;i++) {
    if (!tryInitPort(i)) {
      ports++;
    }
  }
  return (ports > 0) ? 0 : -1;
}

static void __exit tsuart_exit(void)
{
  // This module can be safely unloaded after initialization is complete.
  // We allow module to load only so error conditions can be detected.
}

module_init(tsuart_init);
module_exit(tsuart_exit);

EXPORT_NO_SYMBOLS;
