/*
 *  linux/drivers/serial/tsuart-mwarm.c
 *
 *  TS-UART loader for Technologic Systems TS-MWARM
 *
 *  (c) 2006 Technologic Systems
 *
 * This source code is hereby licensed to the Linux 2.4 kernel under the
 * GPL, and to Technologic Systems customers under the BSD license.
 *
 * 2007-3-23:mos:initial version completed
 *
 */
#include <linux/version.h>
#include <linux/slab.h>
#include <linux/config.h>
#include <linux/module.h>
#include <linux/tty.h>
#include <linux/tty_flip.h>
#include <linux/ioport.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/serial.h>
#include <linux/console.h>
#include <linux/sysrq.h>
#include <linux/proc_fs.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <linux/serial_core.h>
//#include <serial_core.h>
#include <asm/hardware/serial_amba.h>
#include <linux/ts_sbc.h>
#include <asm/uaccess.h>
#include "tsuart.h"

#define TS_MODEL_MWARM 2
#define BOARD_ID_MWARM 0x7261
#define DEBUG_INTERFACE 8

static inline int get_status(void) {
//  return 0;
  return (inb(TS7XXX_IO8_BASE + 0x800000));
}

static inline void set_status(int i) {
  if (i) {
    outb(8|(inb(TS7XXX_IO8_BASE + 0x2000000)),TS7XXX_IO8_BASE + 0x2000000);
  } else {
    outb(~8&(inb(TS7XXX_IO8_BASE + 0x2000000)),TS7XXX_IO8_BASE + 0x2000000);
  }
}

static int mdmctrl(struct tsuart_port *ts,int mctrl) {
  int status;
  int ret = 0;

  if (mctrl & 0x80000000) {
    if (mctrl & TIOCM_DTR) {
      set_status(0);
    } else {
      set_status(1);
    }
  }
  status = get_status();
  if (status & (1 << 2)) {
    ret |= TIOCM_DTR;
  }
  if (status & (1 << 5)) {
    ret |= TIOCM_DSR;
  }
  if (status & (1 << 7)) {
    ret |= TIOCM_CAR;
  }
  return ret;
}

static int tryInitPort(int base) {
  struct tsuart_port *ts;

  ts = kmalloc(sizeof(struct tsuart_port),GFP_KERNEL);
  if (!ts) {
    printk("Unable to allocate memory for tsuart port\n");
    return 0;
  }

  init_tsuart_port(ts);
  ts->u.iotype = SERIAL_IO_MEM;
  ts->get = tsuartGetMEM8;
  ts->put = tsuartPutMEM8;
  ts->mdmctrl = mdmctrl;
  ts->debugFlags = portDebug;
  ts->u.mapbase = /*0 x72000000 + */ base;
  ts->u.irq = 32;
  ts->localStatus |= PORT_STATUS_SHIRQ;
  ts->boardId = BOARD_ID_MWARM;

  if (tsuart_register_port(0,ts) != 0) {
    return 0;
  }
  return 1;
}

static int __init tsuart_init(void)
{
  int ports=0;
	
  if (get_ts_model() != TS_MODEL_MWARM) {
    printk("TS-UART/MWARM did not detect a TS-MWARM board\n");
    return -1;
  } else {
    printk("TS-UART/MWARM detected a TS-MWARM board\n");
  }

  if (tryInitPort(0x12400000)) {
    ports++;
  }
  return (ports > 0) ? 0 : -1;
}

static void __exit tsuart_exit(void)
{
}

module_init(tsuart_init);
module_exit(tsuart_exit);

EXPORT_NO_SYMBOLS;

