/*
 *  linux/drivers/serial/tsuart7400.c
 *
 *  TS-UART loader for Technologic Systems TS-7400
 *
 *  (c) 2006 Technologic Systems
 *
 * This source code is hereby licensed to the Linux 2.4 kernel under the
 * GPL, and to Technologic Systems customers under the BSD license.
 *
 * 2006-05-12:mos:initial version completed
 *
 */
#include <linux/version.h>
#include <linux/slab.h>
#include <linux/config.h>
#include <linux/module.h>
#include <linux/tty.h>
#include <linux/tty_flip.h>
#include <linux/ioport.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/serial.h>
#include <linux/console.h>
#include <linux/sysrq.h>
#include <linux/proc_fs.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <linux/serial_core.h>
//#include <serial_core.h>
#include <asm/hardware/serial_amba.h>
#include <linux/ts_sbc.h>
#include <asm/uaccess.h>
#include "tsuart.h"
#include "ts7412.h"

static int mdmctrl(struct tsuart_port *ts,int mctrl) {
  int ctrl=mctrl;

  if (mctrl & 0x80000000) {
    mdmctrl_dtr_set((mctrl & TIOCM_DTR)?1:0);
    mdmctrl_rts_set((mctrl & TIOCM_RTS)?1:0);
  } else {
    if (mdmctrl_dcd_get()) {
      ctrl |= TIOCM_CAR;
    }
    if (mdmctrl_cts_get()) {
      ctrl |= TIOCM_CTS;
    }
  }
  return (TIOCM_DSR | ctrl);
}

#define BOARD_ID_7400 0x7400
static int tryInitPort(int base) {
  struct tsuart_port *ts;

  ts = kmalloc(sizeof(struct tsuart_port),GFP_KERNEL);
  if (!ts) {
    printk("Unable to allocate memory for tsuart port\n");
    return 0;
  }

  init_tsuart_port(ts);
  ts->u.iotype = SERIAL_IO_MEM;
  ts->get = tsuartGetMEM8;
  ts->put = tsuartPutMEM8;
  ts->mdmctrl = (gpio == 0) ? 0 : mdmctrl;
  ts->debugFlags = portDebug;
  ts->u.mapbase = /*0 x72000000 + */ base;
  ts->u.irq = 33;
  ts->localStatus |= PORT_STATUS_SHIRQ;
  ts->boardId = BOARD_ID_7400;

  if (tsuart_register_port(0,ts) != 0) {
    return 0;
  }
  return 1;
}

static int __init tsuart_init(void)
{
  int ports=0;
	
  if (get_ts_model() != TS_MODEL_7400) {
    printk("TS-UART/7400 did not detect a TS-7400 board\n");
    return -1;
  } 
  if (ts7412_detect_test(0)) {
    printk("TS-UART/7400 detected a TS-7400 with TS-7412 present\n");
  } else {
    printk("TS-UART/7400 detected a TS-7400 board\n");
  }
  if (tryInitPort(0x12400000)) {
    ports++;
  }
  return (ports > 0) ? 0 : -1;
}

static void __exit tsuart_exit(void)
{
}

module_init(tsuart_init);
module_exit(tsuart_exit);

EXPORT_NO_SYMBOLS;

