/*
 *  linux/drivers/serial/tsuart940.c
 *
 *  TS-UART loader for Technologic Systems TS-940
 *
 *  (c) 2006 Technologic Systems
 *
 * This source code is hereby licensed to the Linux 2.4 kernel under the
 * GPL, and to Technologic Systems customers under the BSD license.
 *
 * 2006-05-12:mos:initial version completed
 *
 */
#include <linux/version.h>
#include <linux/slab.h>
#include <linux/config.h>
#include <linux/module.h>
#include <linux/tty.h>
#include <linux/tty_flip.h>
#include <linux/ioport.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/serial.h>
#include <linux/console.h>
#include <linux/sysrq.h>
#include <linux/proc_fs.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <linux/serial_core.h>
//#include <serial_core.h>
#include <asm/hardware/serial_amba.h>
#include <linux/ts_sbc.h>
#include <asm/uaccess.h>
#include "tsuart.h"

#define BOARD_ID_940 0x940

static int tryInitPort(int iobase) {
  struct tsuart_port *ts,*ts2;
  unsigned val;

  ts = kmalloc(sizeof(struct tsuart_port),GFP_KERNEL);
  if (!ts) {
    printk("Unable to allocate memory for tsuart port\n");
    return 0;
  }
  ts2 = kmalloc(sizeof(struct tsuart_port),GFP_KERNEL);
  if (!ts2) {
    kfree(ts);
    printk("Unable to allocate memory for tsuart port\n");
    return 0;
  }

  init_tsuart_port(ts);
  init_tsuart_port(ts2);
  ts->u.iotype = SERIAL_IO_PORT;
  ts->get = tsuartGetIO8;
  ts->put = tsuartPutIO8;
  ts->debugFlags = portDebug;
  ts->u.iobase = TS7XXX_IO8_BASE + TS7XXX_ISAIO + iobase;
  ts->u.irq = 33;
  ts->mdmctrl = 0;
  ts->localStatus |= PORT_STATUS_SHIRQ;
  ts->boardId = BOARD_ID_940;

  ts2->u.iotype = SERIAL_IO_PORT;
  ts2->get = tsuartGetIO8;
  ts2->put = tsuartPutIO8;
  ts2->debugFlags = portDebug;
  ts2->u.iobase = TS7XXX_IO8_BASE + TS7XXX_ISAIO + iobase;
  ts2->u.irq = 33;
  ts2->mdmctrl = 0;
  ts2->localStatus |= PORT_STATUS_SHIRQ;
  ts2->boardId = BOARD_ID_940;

  val = tsuartGetIO8(ts,0);
  if (val == 0x3C) {
    val = tsuartGetIO8(ts,1);
    ts->u.iobase += 0x0C; // point to first STAT register
    ts2->u.iobase += 0x0E; // point to second STAT register
    printk("TS-UART/940 found at port 0x%04X\n",iobase);
    if (tsuart_register_port(0,ts) != 0) {
      return 0;
    }
    if (tsuart_register_port(0,ts2) == 0) {
      return 1;
    }
  }
  return 0;
}

static int __init tsuart_init(void)
{
  int ports=0;

  if (tryInitPort(0x100)) {
    ports++;
  }
  if (tryInitPort(0x110)) {
    ports++;
  }
  if (tryInitPort(0x120)) {
    ports++;
  }
  if (tryInitPort(0x130)) {
    ports++;
  }
  if (ports == 0) {
    printk("TS-UART/940 did not detect a TS-940 board\n");
  }
  return (ports > 0) ? 0 : -1;
}

static void __exit tsuart_exit(void)
{
  // This module can be safely unloaded after initialization is complete.
  // We allow module to load only so error conditions can be detected. 
}

module_init(tsuart_init);
module_exit(tsuart_exit);

EXPORT_NO_SYMBOLS;
